"use strict";
/*
  This program and the accompanying materials are
  made available under the terms of the Eclipse Public License v2.0 which accompanies
  this distribution, and is available at https://www.eclipse.org/legal/epl-v20.html
  
  SPDX-License-Identifier: EPL-2.0
  
  Copyright Contributors to the Zowe Project.
*/
Object.defineProperty(exports, "__esModule", { value: true });
// https://console.spec.whatwg.org/#logger
// consider formatting ideas
// consider grouping ideas 
// time/data/functionname/linenumber
// maybe polyfill from https://www.stacktracejs.com/#!/docs/stacktrace-js 
var LogLevel;
(function (LogLevel) {
    LogLevel[LogLevel["SEVERE"] = 0] = "SEVERE";
    LogLevel[LogLevel["WARNING"] = 1] = "WARNING";
    LogLevel[LogLevel["INFO"] = 2] = "INFO";
    LogLevel[LogLevel["FINE"] = 3] = "FINE";
    LogLevel[LogLevel["FINER"] = 4] = "FINER";
    LogLevel[LogLevel["FINEST"] = 5] = "FINEST";
})(LogLevel = exports.LogLevel || (exports.LogLevel = {}));
var ComponentLogger = /** @class */ (function () {
    function ComponentLogger(parentLogger, componentName) {
        this.parentLogger = parentLogger;
        this.componentName = componentName;
        this.SEVERE = LogLevel.SEVERE;
        this.WARNING = LogLevel.WARNING;
        this.INFO = LogLevel.INFO;
        this.FINE = LogLevel.FINE;
        this.FINER = LogLevel.FINER;
        this.FINEST = LogLevel.FINEST;
    }
    ComponentLogger.prototype.makeSublogger = function (componentNameSuffix) {
        return new ComponentLogger(this.parentLogger, this.componentName + ':' + componentNameSuffix);
    };
    ComponentLogger.prototype.log = function (minimumLevel) {
        var loggableItems = [];
        for (var _i = 1; _i < arguments.length; _i++) {
            loggableItems[_i - 1] = arguments[_i];
        }
        var _a;
        (_a = this.parentLogger).log.apply(_a, [this.componentName, minimumLevel].concat(loggableItems));
    };
    ComponentLogger.prototype.severe = function () {
        var loggableItems = [];
        for (var _i = 0; _i < arguments.length; _i++) {
            loggableItems[_i] = arguments[_i];
        }
        var _a;
        (_a = this.parentLogger).log.apply(_a, [this.componentName, LogLevel.SEVERE].concat(loggableItems));
    };
    ComponentLogger.prototype.info = function () {
        var loggableItems = [];
        for (var _i = 0; _i < arguments.length; _i++) {
            loggableItems[_i] = arguments[_i];
        }
        var _a;
        (_a = this.parentLogger).log.apply(_a, [this.componentName, Logger.INFO].concat(loggableItems));
    };
    ComponentLogger.prototype.warn = function () {
        var loggableItems = [];
        for (var _i = 0; _i < arguments.length; _i++) {
            loggableItems[_i] = arguments[_i];
        }
        var _a;
        (_a = this.parentLogger).log.apply(_a, [this.componentName, Logger.WARNING].concat(loggableItems));
    };
    ComponentLogger.prototype.debug = function () {
        var loggableItems = [];
        for (var _i = 0; _i < arguments.length; _i++) {
            loggableItems[_i] = arguments[_i];
        }
        var _a;
        (_a = this.parentLogger).log.apply(_a, [this.componentName, Logger.FINE].concat(loggableItems));
    };
    return ComponentLogger;
}());
exports.ComponentLogger = ComponentLogger;
var RegExpLevel = /** @class */ (function () {
    function RegExpLevel(regex, level) {
        this.regex = regex;
        this.level = level;
    }
    return RegExpLevel;
}());
var Logger = /** @class */ (function () {
    function Logger() {
        this.componentLoggers = new Map();
        this.knownComponentNames = [];
        this.configuration = {};
        this.destinations = new Array();
        this.previousPatterns = new Array();
    }
    Logger.prototype.addDestination = function (destinationCallback) {
        this.destinations.push(destinationCallback);
    };
    Logger.prototype.shouldLogInternal = function (componentName, level) {
        var configuredLevel = this.configuration[componentName];
        if (configuredLevel === undefined) {
            configuredLevel = Logger.INFO;
        }
        return configuredLevel >= level;
    };
    ;
    Logger.prototype.consoleLogInternal = function (componentName, minimumLevel, prependDate, prependName, prependLevel) {
        var loggableItems = [];
        for (var _i = 5; _i < arguments.length; _i++) {
            loggableItems[_i - 5] = arguments[_i];
        }
        var formatting = '[';
        if (prependDate) {
            var d = new Date();
            var msOffset = d.getTimezoneOffset() * 60000;
            d.setTime(d.getTime() - msOffset);
            var dateString = d.toISOString();
            dateString = dateString.substring(0, dateString.length - 1).replace('T', ' ');
            formatting += dateString + ' ';
        }
        if (prependName) {
            formatting += componentName + ' ';
        }
        if (prependLevel) {
            formatting += LogLevel[minimumLevel];
        }
        formatting += "] -";
        console.log.apply(console, [formatting].concat(loggableItems));
    };
    ;
    Logger.prototype.makeDefaultDestination = function (prependDate, prependName, prependLevel) {
        var theLogger = this;
        return function (componentName, minimumLevel) {
            var loggableItems = [];
            for (var _i = 2; _i < arguments.length; _i++) {
                loggableItems[_i - 2] = arguments[_i];
            }
            if (theLogger.shouldLogInternal(componentName, minimumLevel)) {
                theLogger.consoleLogInternal.apply(theLogger, [componentName, minimumLevel, prependDate, prependName, prependLevel].concat(loggableItems));
            }
        };
    };
    ;
    Logger.prototype.log = function (componentName, minimumLevel) {
        var loggableItems = [];
        for (var _i = 2; _i < arguments.length; _i++) {
            loggableItems[_i - 2] = arguments[_i];
        }
        this.noteComponentNameInternal(componentName);
        this.destinations.forEach(function (destinationCallback) {
            destinationCallback.apply(void 0, [componentName, minimumLevel].concat(loggableItems));
        });
    };
    ;
    Logger.prototype.setLogLevelForComponentPattern = function (componentNamePattern, level) {
        var theLogger = this;
        var componentNameArray = Object.keys(this.configuration);
        var regex = new RegExp(componentNamePattern);
        this.previousPatterns.push(new RegExpLevel(regex, level));
        componentNameArray.filter(function (componentName) {
            return regex.test(componentName);
        }).forEach(function (componentName) {
            theLogger.configuration[componentName] = level;
        });
    };
    ;
    Logger.prototype.setLogLevelForComponentName = function (componentName, level) {
        if (level >= LogLevel.SEVERE && level <= LogLevel.FINEST) {
            this.configuration[componentName] = level;
        }
    };
    Logger.prototype.getComponentLevel = function (componentName) {
        return this.configuration[componentName];
    };
    Logger.prototype.noteComponentNameInternal = function (componentName) {
        if (!this.knownComponentNames.find(function (name) { return name == componentName; })) {
            this.knownComponentNames.push(componentName);
        }
    };
    ;
    Logger.prototype.replayPatternsOnLogger = function (componentName) {
        for (var i = this.previousPatterns.length - 1; i > -1; i--) {
            var pattern = this.previousPatterns[i];
            if (pattern.regex.test(componentName)) {
                this.setLogLevelForComponentName(componentName, pattern.level);
                return true;
            }
        }
        return false;
    };
    Logger.prototype.makeComponentLogger = function (componentName) {
        var componentLogger = this.componentLoggers.get(componentName);
        if (componentLogger) {
            this.consoleLogInternal("<internal>", LogLevel.WARNING, true, false, true, 'Logger created with identical component name to pre-existing logger. Messages overlap may occur.');
        }
        else {
            componentLogger = new ComponentLogger(this, componentName);
            this.configuration[componentName] = LogLevel.INFO;
            this.componentLoggers.set(componentName, componentLogger);
            this.replayPatternsOnLogger(componentName);
        }
        return componentLogger;
    };
    Logger.SEVERE = LogLevel.SEVERE;
    Logger.WARNING = LogLevel.WARNING;
    Logger.INFO = LogLevel.INFO;
    Logger.FINE = LogLevel.FINE;
    Logger.FINER = LogLevel.FINER;
    Logger.FINEST = LogLevel.FINEST;
    return Logger;
}());
exports.Logger = Logger;
/*
  This program and the accompanying materials are
  made available under the terms of the Eclipse Public License v2.0 which accompanies
  this distribution, and is available at https://www.eclipse.org/legal/epl-v20.html
  
  SPDX-License-Identifier: EPL-2.0
  
  Copyright Contributors to the Zowe Project.
*/
//# sourceMappingURL=logger.js.map